import hashlib
import os
import platform

from ctools import cjson, sm_tools, work_path

MACHINE_KEY = b'EnrGffoorbFyTYoS0902YyT1Fhehj4InpbezIDUuPOg='

class MachineInfo:
  machine_code = None

def get_user():
  import getpass
  return getpass.getuser()

def get_machine_code():
  if MachineInfo.machine_code: return MachineInfo.machine_code
  destPath = os.path.join(work_path.get_user_work_path(), "AppData/Local/machine")
  machine_file = os.path.join(destPath, 'machine_code.mc')
  origin_machine_code = get_origin_machine_code()
  if os.path.exists(machine_file):
    with open(machine_file, 'r') as f:
      file_code = f.readline()
      dec_code = ''
      try:
        dec_code = cjson.loads(sm_tools.decrypt_with_sm4(MACHINE_KEY, file_code))
        origin_code = dec_code.get("origin_code")
        if origin_code == origin_machine_code:
          MachineInfo.machine_code = dec_code.get("hash_code")
          print("use current machine code {}".format(MachineInfo.machine_code))
          return MachineInfo.machine_code
      except Exception:
        print('machine code file is error: {} {}'.format(file_code, dec_code))
  hash_machine_code = get_hash_machine_code(origin_machine_code)
  machine_code = {"origin_code": origin_machine_code, "hash_code": hash_machine_code}
  enc_code = sm_tools.encrypt_with_sm4(MACHINE_KEY, cjson.dumps(machine_code))
  os.makedirs(destPath, exist_ok=True)
  with open(machine_file, 'w') as f:
    f.write(enc_code)
    f.flush()
  MachineInfo.machine_code = hash_machine_code
  print("init new machine code {}".format(hash_machine_code))
  return MachineInfo.machine_code


def get_origin_machine_code():
  # 获取CPU序列号
  cpu_serial = platform.processor()
  # 获取硬盘序列号
  disk_serial = ''
  if platform.system() == 'Windows':
    import ctypes
    kernel32 = ctypes.windll.kernel32
    volume_serial = ctypes.c_ulonglong(0)
    kernel32.GetVolumeInformationW(
      ctypes.c_wchar_p("C:\\"),
      None,
      0,
      ctypes.pointer(volume_serial),
      None,
      None,
      None,
      0
    )
    disk_serial = str(volume_serial.value)
  combined_info = cpu_serial + disk_serial + '-gomyck'
  return hashlib.md5(combined_info.encode()).hexdigest()


def get_hash_machine_code(origin_code):
  import uuid
  code = origin_code + uuid.uuid1().hex
  machine_code = hashlib.md5(code.encode()).hexdigest()
  return machine_code.upper()

def get_public_ip():
  import requests
  try:
    response = requests.get("https://api.ipify.org?format=json")
    ip = response.json()["ip"]
    return ip
  except Exception as e:
    return f"Failed to get public IP: {e}"

def get_local_ipv4():
  import psutil
  import socket
  interfaces = psutil.net_if_addrs()
  for interface, addresses in interfaces.items():
    for address in addresses:
      if address.family == socket.AF_INET and not address.address.startswith("127."):
        return address.address
  print("Failed to get local IPv4 address, try another way...")
  s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
  try:
    s.connect(("8.8.8.8", 80))
    ip = s.getsockname()[0]
  except Exception:
    ip = '127.0.0.1'
  finally:
    s.close()
  return ip

def get_remote_ipv4():
  from bottle import request
  try:
    return request.remote_route[0]
  except:
    return '127.0.0.1'

def get_proc_pid_by(cmdline):
  import psutil
  """
  根据命令行信息获取进程pid
  :param cmdline:
  :return:
  """
  pid_list = []
  proc_list = psutil.process_iter()
  for proc in proc_list:
    try:
      cmdline_str = "".join(proc.cmdline())
      if cmdline in cmdline_str:
        pid_list.append(proc.pid)
    except Exception:
      pass
  return pid_list

def get_os_architecture():
  if '64' in platform.machine():
    return '64'
  else:
    return '32'


def get_os_architecture4x():
  if '64' in platform.machine():
    return 'x64'
  else:
    return 'x86'


def get_os_version():
  system = platform.system()
  if system == "Windows":
    version = platform.win32_ver()[0]
    return version
  else:
    return "Unsupported OS"


def get_platform_name():
  version = get_os_version()
  if version == '7':
    return 'win{}{}'.format(version, get_os_architecture4x())
  else:
    return 'win{}'.format(version)



import inspect
import os
import sys


def get_current_path(subPath: str = '') -> str:
  """
  获取当前文件所在路径
  :param subPath: 子路径
  :return: 当前文件所在路径
  """
  if getattr(sys, 'frozen', False):
    base_path = sys._MEIPASS
  else:
    caller_frame = inspect.currentframe().f_back
    caller_path = caller_frame.f_globals["__file__"]
    base_path = os.path.dirname(caller_path)
  return os.path.join(base_path, subPath)


def get_app_path(subPath: str = '') -> str:
  """
  获取应用程序工作时所在路径
  :param subPath: 子路径
  :return: 应用程序工作时所在路径
  """
  if getattr(sys, 'frozen', False):
    base_path = sys._MEIPASS
  else:
    base_path = os.getcwd()
    while not os.path.isfile(os.path.join(base_path, 'README.md')):
      base_path = os.path.dirname(base_path)
  return os.path.join(base_path, subPath)


def get_user_work_path(subPath: str = '', mkdir: bool = False):
  """
  获取用户工作目录
  :param subPath: 拼接的子路径
  :param mkdir: 是否创建目录
  :return: 用户工作目录
  """
  path = os.path.join(os.path.expanduser("~"), subPath)
  if mkdir and not os.path.exists(path): os.makedirs(path, exist_ok=True)
  return path


def get_user_temp_path(subPath: str = '') -> str:
  """
  获取临时文件夹所在路径
  :return: 临时文件夹所在路径
  """
  return os.path.join(os.environ.get('TEMP'), subPath)


def get_Users_path(subPath: str = '') -> str:
  """
  获取用户目录所在路径 C:/Users
  :return: 用户目录所在路径
  """
  return os.path.join(os.path.dirname(os.path.expanduser("~")), subPath)


def get_install_path(subPath: str = '') -> str:
  """
  获取安装包安装的路径
  :return: 安装包安装的路径
  """
  return os.path.join(os.getcwd(), subPath)



import os
import time

from business.common.constant import Scheduler
from lxml import etree
from lxml import html
from pynput import keyboard
from selenium import webdriver
from selenium.common import NoSuchElementException, WebDriverException, NoSuchWindowException
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from urllib3.exceptions import MaxRetryError

from ctools import application, string_tools

keyboard_listener = None
ctrl_pressed = None
g_driver = None
g_callback = None
g_result = []
g_quite_flag = False
picture_path = ""

def get_hovered_element_html(url, explore, callback):
  global g_driver, g_callback, g_result
  g_callback = callback
  g_driver = explore.init()
  driver = g_driver
  driver.maximize_window()
  driver.get(url)
  start_keyboard_listener()
  handle_arr = [driver.current_window_handle]
  close_page_flag = False
  while g_driver and not g_quite_flag:
    try:
      try:
        if len(driver.window_handles) == 0:
          break_func()
          break
      except WebDriverException:
        break_func()
        break
      except MaxRetryError:
        break_func()
        break
      try:
        driver.find_element(value="ck-overlay")
      except NoSuchElementException as e:
        driver.execute_script(explore.listen_script)
      except Exception:
        pass
      for handle in driver.window_handles:
        if handle in handle_arr: continue
        if not close_page_flag: driver.execute_script(overlay_script)
        handle_arr.append(handle)
        driver.switch_to.window(handle)
        try:
          driver.execute_script(explore.listen_script)
        except Exception:
          continue
      close_page_flag = False
      time.sleep(0.5)
      cursor_html = driver.execute_script(explore.return_script)
      if cursor_html:
        g_result.clear()
        cursor_dom = html.fragment_fromstring(cursor_html)
        cursor_dom.set("ck-flag", "ck")
        enhance_cursor_html = etree.tostring(cursor_dom, method="html", encoding="unicode")
        page_html = driver.page_source.replace(cursor_html, enhance_cursor_html)
        page_dom_tree = etree.ElementTree(etree.HTML(page_html))
        match_dom = page_dom_tree.xpath('//*[@ck-flag="ck"]')
        for ck_element in match_dom:
          ck_xpath = page_dom_tree.getpath(ck_element)
          #print('XPATH IS {}'.format(ck_xpath))
          try:
            ele = driver.find_element(By.XPATH, ck_xpath)
            #print('XPATH_HTML: {}'.format(ele.get_attribute('outerHTML')))
          except Exception:
            pass
          g_result.append(ck_xpath)
    except NoSuchWindowException as e:
      close_page_flag = True
      handle_arr = []
      try:
        driver.switch_to.window(driver.window_handles[-1])
        driver.execute_script(hide_shade)
      except Exception:
        print('切换HANDLE失败:', e)
        break_func()
        break
    except Exception as e:
      print('全局错误:', e)
      break_func()
      break
  try:
    g_driver.quit()
  except Exception as e:
    pass

def break_func():
  keyboard_listener.stop()
  g_callback(None)

def on_press(key):
  global keyboard_listener, ctrl_pressed, g_driver, g_quite_flag, picture_path
  if key == keyboard.Key.ctrl_l and not ctrl_pressed:
    ctrl_pressed = True
  elif key == keyboard.Key.esc:
    g_result.clear()
    keyboard_listener.stop()
    g_quite_flag = True
    g_callback(None)
  elif hasattr(key, 'vk') and key.vk == 192 and ctrl_pressed:
    # print("g_result: %s" % g_result)
    if g_result:
      try:
        select_element = g_driver.find_element(By.XPATH, g_result[0])
        picture_path = "%s.png" % os.path.join(application.Server.screenshotPath, "element-"+string_tools.get_uuid())
        select_element.screenshot(picture_path)
      except Exception:
        pass

    keyboard_listener.stop()
    g_quite_flag = True
    g_callback(g_result)

def on_release(key):
  global ctrl_pressed
  if key == keyboard.Key.ctrl_l:
    ctrl_pressed = False

def start_keyboard_listener():
  global keyboard_listener
  keyboard_listener = keyboard.Listener(on_press=on_press, on_release=on_release)
  keyboard_listener.start()


overlay_script = """
// 创建遮罩层
var shade = document.createElement('div');
shade.id = 'ck-shade-parent';
shade.style.position = 'fixed';
shade.style.top = '0';
shade.style.left = '0';
shade.style.width = '100%';
shade.style.height = '100%';
shade.style.backgroundColor = '#000'; // 使用纯色背景
shade.style.filter = 'alpha(opacity=60)'; // 设置透明度，IE8-IE9
shade.style.opacity = '0.8'; // 设置透明度，现代浏览器
shade.style.zIndex = '9999';
document.body.appendChild(shade);

// 创建覆盖内容
var overlayContent = document.createElement('div');
overlayContent.id = 'ck-shade-oc';
overlayContent.className = 'overlay-content';
overlayContent.style.position = 'absolute';
overlayContent.style.top = '50%';
overlayContent.style.left = '50%';
overlayContent.style.transform = 'translate(-50%, -50%)';
overlayContent.style.backgroundColor = 'white';
overlayContent.style.padding = '20px';
overlayContent.style.borderRadius = '8px';
overlayContent.style.textAlign = 'center';
shade.appendChild(overlayContent);

// 创建消息内容
var message = document.createElement('p');
message.id = 'ck-shade-msg';
message.innerText = '当前页面未激活，请关闭激活状态的录制页面';
message.style.color = '#000'; // 设置文本颜色为黑色
message.style.fontSize = '16px'; // 设置文本大小
overlayContent.appendChild(message);

"""

hide_shade = """
var shade = document.getElementById('ck-shade-parent');
if (shade) {
    shade.parentNode.removeChild(shade);
}
"""

class IE:

  @staticmethod
  def init():
    ie_options = webdriver.IeOptions()
    ie_options.ignore_protected_mode_settings = True
    ie_options.ignore_zoom_level = True
    ie_options.require_window_focus = True
    return webdriver.Ie(options=ie_options)

  listen_script = """
    var overlay = document.createElement('div');
    overlay.id = 'ck-overlay';
    overlay.style.position = 'absolute';
    overlay.style.border = '2px solid red';
    overlay.style.pointerEvents = 'none';
    overlay.style.zIndex = '9999';
    document.body.appendChild(overlay);
    var addEvent = function(elem, type, eventHandle) {
      if (elem == null || typeof(elem) == 'undefined') return;
      if (elem.addEventListener) {
        elem.addEventListener(type, eventHandle, false);
      } else if (elem.attachEvent) {
        elem.attachEvent('on' + type, eventHandle);
      } else {
        elem['on' + type] = eventHandle;
      }
    };

    addEvent(document, 'mousemove', function(e) {
      e = e || window.event;
      var element = document.elementFromPoint(e.clientX, e.clientY);
      window.hoveredElement = element;
      if (element === overlay) return;
      var rect = element.getBoundingClientRect();
      overlay.style.left = (rect.left + (window.pageXOffset || document.documentElement.scrollLeft || document.body.scrollLeft) - 5) + 'px';
      overlay.style.top = (rect.top + (window.pageYOffset || document.documentElement.scrollTop || document.body.scrollTop) - 5) + 'px';
      overlay.style.width = (element.offsetWidth + 10)  + 'px';
      overlay.style.height = (element.offsetHeight + 10) + 'px';
    });
    """

  return_script = """
    return window.hoveredElement ? window.hoveredElement.outerHTML : null;
  """

  overlay_script = overlay_script


class Chrome:

  @staticmethod
  def init():
    option = Options()
    option.binary_location = Scheduler.CHROME_PATH
    service = Service(Scheduler.CHROME_DRIVER_PATH)
    return webdriver.Chrome(options=option, service=service)

  listen_script = """
    var overlay = document.createElement('div');
    overlay.id = 'ck-overlay';
    overlay.style.position = 'absolute';
    overlay.style.border = '2px solid red';
    overlay.style.pointerEvents = 'none';
    overlay.style.zIndex = '9998';
    document.body.appendChild(overlay);

    function throttle(func, limit) {
      let inThrottle;
      return function() {
        const args = arguments,
          context = this;
        if (!inThrottle) {
          func.apply(context, args);
          inThrottle = true;
          setTimeout(() => inThrottle = false, limit);
        }
      };
    }

    document.addEventListener('mousemove', throttle(function(e) {
      var element = document.elementFromPoint(e.clientX, e.clientY);
      window.hoveredElement = element;
      if(element.id.indexOf('ck-shade') !== -1) return;
      if (window.hoveredElement) {
        var rect = element.getBoundingClientRect();
        overlay.style.left = (rect.left + window.pageXOffset - 5) + 'px';
        overlay.style.top = (rect.top + window.pageYOffset - 5) + 'px';
        overlay.style.width = (element.offsetWidth + 10) + 'px';
        overlay.style.height = (element.offsetHeight + 10) + 'px';
      }
    }, 50));
    """

  return_script = """
    return window.hoveredElement ? window.hoveredElement.outerHTML : null;
  """

  overlay_script = overlay_script


def callback(xpath):
  print("Hovered Element XPATH IS :", xpath)

def get_element(url: str = None, explore: str = "chrome"):
  global keyboard_listener, ctrl_pressed, g_driver, g_callback, g_result, g_quite_flag, picture_path
  keyboard_listener = None
  ctrl_pressed = None
  g_driver = None
  g_callback = None
  g_result = []
  g_quite_flag = False
  picture_path = ""

  if explore == "chrome":
    explore = Chrome
  else:
    explore = IE

  if "http" not in url[:5]:
    url = "http://%s" % url

  get_hovered_element_html(url, explore, callback)

  return g_result, picture_path

if __name__ == "__main__":
  # from explore_record_core import get_hovered_element_html, Chrome
  g_result, picture_path = get_element("weibo.com")
  print(g_result, picture_path)



import os
from typing import List

import xlrd
import xlwt
from docx import Document
from docx.enum.table import WD_TABLE_ALIGNMENT, WD_ALIGN_VERTICAL
from docx.oxml import OxmlElement, parse_xml
from docx.oxml.ns import qn, nsdecls
from docx.shared import Pt, RGBColor
from docxtpl import DocxTemplate, RichText
from openpyxl import load_workbook
from openpyxl.styles import Side, Border
from xlstpl import writer as xls_writer
from xlsxtpl import writerx as xlsx_writerx
from xlutils.copy import copy

from ctools import cjson
from ctools.word_fill_entity import WordTable, WordCell, Style, Font, Alignment

EXCEL_WORD_ALIGNMENT = {"left": WD_TABLE_ALIGNMENT.LEFT,
                        "center": WD_TABLE_ALIGNMENT.CENTER,
                        "right": WD_TABLE_ALIGNMENT.RIGHT}
EXCEL_WORD_VERTICAL = {"top": WD_ALIGN_VERTICAL.TOP,
                       "center": WD_ALIGN_VERTICAL.CENTER,
                       "bottom": WD_ALIGN_VERTICAL.BOTTOM}


class DocxTools:

  @staticmethod
  def get_docx(path=None):
    return Document(path)

  @staticmethod
  def add_rich_text(wildcard: dict):
    try:
      new_wildcard = {}
      if wildcard:
        for k, v in wildcard.items():
          if "_color" in k:
            n_k = k[:k.index('_color')]
            new_wildcard[n_k] = RichText(wildcard[n_k], color=v)
          elif k not in new_wildcard:
            new_wildcard[k] = v
      return new_wildcard
    except Exception as e:
      print("使用word_fill.add_rich_text函数异常: %s" % e)
    return wildcard

  def tables_format(self, input_path: str, table_data: List[dict] = None):
    if table_data:
      doc = Document(input_path)
      for index, table in enumerate(doc.tables):
        if index <= len(table_data) - 1:
          td = table_data[index]
          start_position = td.get('start_position')
          data = td.get('data')

          for r in range(0, len(data)):
            for c in range(0, len(data[r])):
              row = r + start_position[0]
              col = c + start_position[1]

              for i in range((row - len(table.rows)) + 1):
                table.add_row()
              try:
                table.cell(row, col).text = str(data[r][c])
              except IndexError:
                pass

          self.set_cell_border(table)
      doc.save(input_path)

  def set_cell_border(self, table):
    return self.set_table_boarder(
      table,
      top={"sz": 4, "val": "single", "color": "#000000"},
      bottom={"sz": 4, "val": "single", "color": "#000000"},
      left={"sz": 4, "val": "single", "color": "#000000"},
      right={"sz": 4, "val": "single", "color": "#000000"},
      insideV={"sz": 4, "val": "single", "color": "#000000"},
      insideH={"sz": 4, "val": "single", "color": "#000000"}
    )

  @staticmethod
  def set_table_boarder(table, **kwargs):
    """
    Set table`s border
    Usage:
    set_table_border(
        cell,
        top={"sz": 12, "val": "single", "color": "#FF0000"},
        bottom={"sz": 12, "color": "#00FF00", "val": "single"},
        left={"sz": 24, "val": "dashed"},
        right={"sz": 12, "val": "dashed"},
    )
    """
    borders = OxmlElement('w:tblBorders')
    for tag in ('bottom', 'top', 'left', 'right', 'insideV', 'insideH'):
      edge_data = kwargs.get(tag)
      if edge_data:
        any_border = OxmlElement(f'w:{tag}')
        for key in ["sz", "val", "color", "space", "shadow"]:
          if key in edge_data:
            any_border.set(qn(f'w:{key}'), str(edge_data[key]))
        borders.append(any_border)
        table._tbl.tblPr.append(borders)

  @staticmethod
  def merge_word(document: Document, input_file):
    """
    合并docx文件内容
    :param document:
    :param input_file:
    :return:
    """
    doc = Document(input_file)
    for element in doc.element.body:
      document.element.body.append(element)

  @staticmethod
  def merge_excel(document: Document, excel_path):
    """
    合并excel文件内容
    :param document:
    :param excel_path: excel文件路径
    :return:
    """
    wt = ExcelTools().get_excel_cells(excel_path)
    max_rows = wt.max_rows
    max_cols = wt.max_cols
    cells = wt.cells
    merged_cells = wt.merged_cells

    doc_body = document.element.body
    # 创建Word表格
    table = document.add_table(rows=max_rows, cols=max_cols, style="Table Grid")
    if len(doc_body) > 0:
      doc_body[-1].addnext(table._tbl)
    table.autofit = False

    for cell in cells:
      position = cell.position
      value = cell.value
      style = cell.style

      t_cell = table.cell(position[0], position[1])
      t_cell.text = str(value)
      paragraph = t_cell.paragraphs[0]
      run = paragraph.runs[0]

      if style:
        if style.font:
          run.font.name = style.font.name
          run.font.size = Pt(style.font.size)
          run.font.bold = style.font.bold
          run.font.italic = style.font.italic
          run.font.underline = style.font.underline
          run._element.rPr.rFonts.set(qn('w:eastAsia'), style.font.name)
          if style.font.color:
            run.font.color.rgb = RGBColor.from_string(style.font.color)

        if style.bg_color:
          t_cell._tc.get_or_add_tcPr().append(parse_xml(r'<w:shd {} w:fill="{bgColor}"/>'.format(nsdecls('w'), bgColor=style.bg_color)))

        paragraph.paragraph_format.alignment = EXCEL_WORD_ALIGNMENT.get(style.alignment.horizontal)
        t_cell.vertical_alignment = EXCEL_WORD_VERTICAL.get(style.alignment.vertical)

    # 合并单元格
    for mc in merged_cells:
      first_cell = None
      first_group_cell = []
      for p in mc:
        for i, r in enumerate(p):
          if i == 0:
            first_cell = table.cell(r[0], r[1])
            first_group_cell.append(first_cell)
          else:
            first_cell.merge(table.cell(r[0], r[1]))

      if len(first_group_cell) > 1:
        for i, cell in enumerate(first_group_cell):
          if i == 0:
            first_cell = cell
          else:
            first_cell.merge(cell)


class ExcelTools:

  @staticmethod
  def tables_format_xls(input_path: str, table_data: List[dict] = None):
    if table_data:
      wb = xlrd.open_workbook(input_path, formatting_info=True)
      wb = copy(wb)

      borders = xlwt.Borders()
      borders.left = 1
      borders.right = 1
      borders.top = 1
      borders.bottom = 1

      style = xlwt.XFStyle()
      style.borders = borders

      sheet = wb.get_sheet(0)
      for td in table_data:
        start_position = td.get('start_position')
        data = td.get('data')

        for r in range(0, len(data)):
          for c in range(0, len(data[r])):
            row = r + start_position[0]
            col = c + start_position[1]
            sheet.write(row, col, data[r][c], style=style)

      wb.save(input_path)

  @staticmethod
  def tables_format_xlsx(input_path: str, table_data: List[dict] = None):
    if table_data:
      wb = load_workbook(input_path)
      sheet = wb[wb.sheetnames[0]]

      side = Side(style='thin', color='FF000000')
      border = Border(left=side, right=side, top=side, bottom=side)

      for td in table_data:
        start_position = td.get('start_position')
        data = td.get('data')

        for r in range(0, len(data)):
          for c in range(0, len(data[r])):
            row = r + start_position[0] + 1
            col = c + start_position[1] + 1
            cell = sheet.cell(row=row, column=col)
            cell.value = data[r][c]
            cell.border = border

      wb.save(filename=input_path)

  def get_excel_cells(self, excel_path) -> WordTable:
    """
    获取excel所有有效单元格信息
    :param excel_path:
    :return:
    """
    cells = []
    merged_cells = []
    eff_max_rows, eff_max_cols, real_max_rows, real_max_cols = 0, 0, 0, 0
    suffix = os.path.splitext(excel_path)[-1].lower()
    if suffix == ".xls":
      wb = xlrd.open_workbook(excel_path, formatting_info=True)
      sheet = wb.sheets()[0]

      eff_max_rows, eff_min_rows, eff_max_cols, eff_min_cols, real_max_rows, real_max_cols = self.get_excel_rows_cols(sheet, is_xlsx=False)

      for r in range(0, real_max_rows):
        for c in range(0, real_max_cols):
          val = sheet.cell_value(r, c)
          if val and len(str(val).strip()) > 0:
            position = [r - eff_min_rows, c - eff_min_cols]
            xfx = sheet.cell_xf_index(position[0], position[1])
            xf = wb.xf_list[xfx]

            horizontal = None
            vertical = None
            hor_align = xf.alignment.hor_align
            if hor_align == 1:
              horizontal = "left"
            elif hor_align == 2:
              horizontal = "center"
            elif hor_align == 3:
              horizontal = "right"

            vert_align = xf.alignment.vert_align
            if vert_align == 0:
              vertical = "top"
            elif vert_align == 1:
              vertical = "center"
            elif vert_align == 2:
              vertical = "bottom"

            font = wb.font_list[xf.font_index]

            style = Style()
            font_size = round(font.height / 20)
            if font_size:
              if eff_max_cols < 20:
                font_size = 5 if round(font_size * 0.8) <= 5 else round(font_size * 0.8)
              else:
                font_size = 5 if round(font_size * 0.7) <= 5 else round(font_size * 0.7)
            style.font = Font(font.name, font_size, bool(font.bold), bool(font.italic), bool(font.underlined))
            # style.font.set_color(cell.font.color)
            style.alignment = Alignment(horizontal, vertical)
            # style.set_bg_color(cell.fill.start_color)

            word_cell = WordCell()
            word_cell.position = position
            word_cell.value = str(val)
            word_cell.style = style
            cells.append(word_cell)

      merged_cells = self.get_excel_merged_cells(sheet, eff_min_rows, eff_min_cols, is_xlsx=False)

    elif suffix == ".xlsx":
      wb = load_workbook(excel_path)
      sheet = wb[wb.sheetnames[0]]
      eff_max_rows, eff_min_rows, eff_max_cols, eff_min_cols, real_max_rows, real_max_cols = self.get_excel_rows_cols(sheet)

      for r in range(1, real_max_rows + 1):
        for c in range(1, real_max_cols + 1):
          cell = sheet.cell(r, c)
          val = cell.value
          if val and len(str(val).strip()) > 0:
            position = [r - eff_min_rows, c - eff_min_cols]

            style = Style()
            font_size = cell.font.size
            if font_size:
              if eff_max_cols < 20:
                font_size = 5 if round(font_size * 0.8) <= 5 else round(font_size * 0.8)
              else:
                font_size = 5 if round(font_size * 0.7) <= 5 else round(font_size * 0.7)

            style.font = Font(cell.font.name, font_size, cell.font.bold, cell.font.italic, cell.font.underline)
            style.font.set_color(cell.font.color)
            vertical = cell.alignment.vertical
            if cell.alignment.vertical is None:
              vertical = "bottom"
            style.alignment = Alignment(cell.alignment.horizontal, vertical)
            style.set_bg_color(cell.fill.start_color)

            word_cell = WordCell()
            word_cell.position = position
            word_cell.value = str(val)
            word_cell.style = style
            cells.append(word_cell)

      merged_cells = self.get_excel_merged_cells(sheet, eff_min_rows, eff_min_cols)

    word_table = WordTable()
    word_table.max_rows = eff_max_rows
    word_table.max_cols = eff_max_cols
    word_table.cells = cells
    word_table.merged_cells = merged_cells
    return word_table

  @staticmethod
  def get_excel_rows_cols(sheet, is_xlsx=True):
    """
    获取excel有效的最大行数和列数
    :param sheet:
    :param is_xlsx:
    :return:
    """
    row_list, col_list = [], []
    if is_xlsx:
      start_index = 1
      real_max_rows, real_max_cols = sheet.max_row, sheet.max_column
      real_max_cols = sheet.max_column
    else:
      start_index = 0
      real_max_rows, real_max_cols = sheet.nrows, sheet.ncols

    index = 0
    eff_max_rows, eff_max_cols = real_max_rows, real_max_cols
    eff_min_rows, eff_min_cols = start_index, start_index
    for r in range(start_index, real_max_rows):
      for c in range(start_index, real_max_cols):
        val = sheet.cell(r, c).value if is_xlsx else sheet.cell_value(r, c)
        if val and len(str(val).strip()) > 0:
          row_list.append(r)
          col_list.append(c)
          if index == 0:
            eff_min_rows = r
            eff_min_cols = c
            index = 1
    if len(row_list) > 0:
      eff_max_rows = max(row_list) - min(row_list) + start_index + 1
    if len(col_list) > 0:
      eff_max_cols = max(col_list) - min(col_list) + start_index + 1
    return eff_max_rows, eff_min_rows, eff_max_cols, eff_min_cols, real_max_rows, real_max_cols

  @staticmethod
  def get_excel_merged_cells(sheet, d_row, d_col, is_xlsx=True):
    """
    获取excel所有合并的单元格信息
    :param sheet:
    :param d_row: 有效起始行和实际起始行的差值
    :param d_col:
    :param is_xlsx:
    :return:
    """
    merged_cells = []
    if is_xlsx:
      merged_ranges = sheet.merged_cells.ranges
      for merged_range in merged_ranges:
        cells = []
        for r in range(merged_range.min_row, merged_range.max_row + 1):
          rows = []
          for c in range(merged_range.min_col, merged_range.max_col + 1):
            rows.append([r - d_row, c - d_col])
          cells.append(rows)
        merged_cells.append(cells)
    else:
      mergeds = sheet.merged_cells
      for merged in mergeds:
        cells = []
        for r in range(merged[0], merged[1]):
          rows = []
          for c in range(merged[2], merged[3]):
            rows.append([r - d_row, c - d_col])
          cells.append(rows)
        merged_cells.append(cells)

    return merged_cells


def template_format(input_path: str, output_path: str, wildcard: dict = None, table_data=None):
  """
  填充模板信息
  :param input_path: 模板文件路径
  :param output_path: 填充后输出文件路径
  :param wildcard: 模板通配符数据
  :param table_data: 模板中表格数据
  :return:
  """
  if os.path.exists(input_path):
    suffix = os.path.splitext(input_path)[-1].lower()
    if table_data and not isinstance(table_data, list):
      table_data = cjson.loads(table_data)

    if suffix == ".xls":
      tpl = xls_writer.BookWriter(input_path)
      tpl.render_book([wildcard])
      tpl.save(output_path)
      ExcelTools().tables_format_xls(output_path, table_data)
    elif suffix == ".xlsx":
      tpl = xlsx_writerx.BookWriter(input_path)
      tpl.render_book([wildcard])
      tpl.save(output_path)
      ExcelTools().tables_format_xlsx(output_path, table_data)
    elif suffix == ".docx":
      tpl = DocxTemplate(input_path)
      wildcard = DocxTools.add_rich_text(wildcard)
      tpl.render(wildcard)
      tpl.save(output_path)
      DocxTools().tables_format(output_path, table_data)


def check_format(input_path: str):
  """
  校验文件后缀与文件格式是否匹配
  :param input_path:
  :return:
  """
  res = True
  try:
    suffix = os.path.splitext(input_path)[-1].lower()
    if suffix == ".xls":
      xls_writer.BookWriter(input_path)
    elif suffix == ".xlsx":
      xlsx_writerx.BookWriter(input_path)
    elif suffix == ".docx":
      DocxTemplate(input_path)
  except Exception:
    res = False
  return res

def excel_optimize(input_path: str):
  """
  优化excel内容清除无法使用的隐患
  :param input_path:
  :return:
  """
  try:
    suffix = os.path.splitext(input_path)[-1].lower()
    if suffix == ".xlsx":
      wb = load_workbook(input_path)
      sheet = wb[wb.sheetnames[0]]
      real_max_rows, real_max_cols = sheet.max_row, sheet.max_column
      is_optimize = False

      if real_max_cols >= 100:
        sheet.delete_cols(100, real_max_cols)
        is_optimize = True

      if is_optimize:
        wb.save(input_path)
  except Exception as e:
    print("优化模板内容清除无法使用的隐患异常: %s" % e)


# 示例用法：从第1行到第5行、从第1列到第3列的区域复制到Word文档中
# excel_file = r'C:\Users\DELL\xxx/xxx-rpa/document\test-2024-04-01_09-05-26.xlsx'       # Excel文件名
# excel_file_1 = r'E:\test\c.xlsx'       # Excel文件名
# excel_file_2 = r'E:\test\b.xlsx'       # Excel文件名
# word_file = r'E:\test\test.docx'     # 输出Word文件名
# # xls_insert_docx(excel_file, word_file, start_row=1, end_row=11, start_col=1, end_col=28)
#
# doc = Document()
# input_files = r"E:\test\样式问题2_11-31-28.docx"
# DocxTools.merge_word(doc, input_files)
# DocxTools.merge_excel(doc, excel_file_1)
# DocxTools.merge_word(doc, input_files)
# DocxTools.merge_excel(doc, excel_file_1)
# DocxTools.merge_excel(doc, excel_file_2)
# DocxTools.merge_excel(doc, excel_file_1)
# DocxTools.merge_word(doc, input_files)
#
# doc.save(word_file)


# input_path = [r"E:\test\a.xlsx", r"E:\test\哈尔滨地铁数据文书模板.docx"]
#
#
# wildcard = {"title": "模板标题", "department": "技术平台部"}
# table_1 = {
#   'start_position': (13, 0),
#   'data': [['111', '112', '113'], ['121', '122', '123']]
# }
#
# table_2 = {
#   'start_position': (20, 0),
#   'data': [['211', '212', '213'], ['221', '222', '223']]
# }
#
# table_3 = {
#   'start_position': (1, 0),
#   'data': [['311', '312', '313'], ['321', '322', '323']]
# }
#
# table = {
#   'start_position': (13, 0),
#   'data': [['111', '112', '113'], ['121', '122', '123']]
# }
# doc_1 = [table]
# doc_2 = [table]
#
# doc_table = [doc_1, doc_2]
#
#
# doc_table_data = cjson.dumps(doc_table)
# rest_data = cjson.loads(doc_table_data)
#
# # print(type(doc_table_data), doc_table_data)
#
# output_dir = r"E:\test\out"
# for i, path in enumerate(input_path):
#   output_path = os.path.join(output_dir, os.path.split(path)[-1])
#   template_format(path, output_path, wildcard=wildcard, doc_table_data=rest_data[i])


# ext = os.path.splitext(input_path)[-1]
# print(check_format(input_path))

# from docxtpl import DocxTemplate, RichText






from bottle import ServerAdapter, Bottle
from geventwebsocket.handler import WebSocketHandler

from ctools import sys_log

"""
module_names = list(globals().keys())
def get_modules():
  mods = []
  for modname in module_names:
    if modname == 'base' or modname == 'online' or modname.startswith('__') or modname == 'importlib': continue
    module = globals()[modname]
    mods.append(module)
  return mods

def get_ws_modules():
  from . import websocket
  return [websocket]
"""

"""
ws_app = bottle_web_base.init_app('/websocket_demo')

@ws_app.route('/script_debug', apply=[websocket])
@bottle_web_base.rule('DOC:DOWNLOAD')
def script_debug(ws: WebSocket):
  print(123)

socket_app = bottle_websocket.init_bottle()
socket_app.mount(app.context_path, ws_app)
socket_app.run()
"""

_default_port = 8887

class CBottle:

  def __init__(self, bottle: Bottle, port=_default_port, quiet=False):
    self.port = port
    self.quiet = quiet
    self.bottle = bottle

  def run(self):
    socket_server = WebSocketServer(port=self.port)
    self.bottle.run(server=socket_server, quiet=self.quiet)

  def mount(self, context_path, app):
    self.bottle.mount(context_path, app)

def init_bottle(port=_default_port, quiet=False) -> CBottle:
  bottle = Bottle()
  return CBottle(bottle, port, quiet)

class CustomWebSocketHandler(WebSocketHandler):
  def log_request(self):
    if '101' not in str(self.status):
      log_msg = self.format_request()
      for nk in sys_log.neglect_keywords:
        if nk in log_msg:
          return
      self.logger.info(log_msg)

class WebSocketServer(ServerAdapter):

  def __init__(self, host='0.0.0.0', port=_default_port):
    super().__init__(host, port)
    self.server = None

  def run(self, handler):
    from gevent import pywsgi
    self.server = pywsgi.WSGIServer((self.host, self.port), handler, handler_class=CustomWebSocketHandler)
    self.server.serve_forever()

  def stop(self):
    self.server.stop()
